//Title:          mc-tos
//Version:        1.0b
//Entered-date:   10 Nov 2008 
//Description:    Sending multicast pkts with tos value
//Author:         Anders Gustavsson, and.gus@hotmail.com
//Maintained-by:  Anders Gustavsson, and.gus@hotmail.com
//Primary-site:   http://sourceforge.net/projects/mctos/
//Platforms:      win/dos
//Copying-policy: GNU GPL

#define _CRT_SECURE_CPP_OVERLOAD_SECURE_NAMES 1

#include <winsock2.h>
#include <ws2tcpip.h>
#include <qos.h>
#include <ntddndis.h>
#include <traffic.h>
#include <qossp.h>
#include <stdio.h>
#include <tcerror.h>	// for traffic control error codes
#include <winnls.h>

#define IPPROTO_UDP 17

#pragma warning(disable : 4995)
#pragma comment(lib, "traffic.lib")
#pragma comment(lib, "Ws2_32.lib")

// Call back function for traffic control. We are not using it, but required for API call.
void MyClNotifyHandler( HANDLE ClRegCtx,HANDLE ClIfcCtx,ULONG Event,HANDLE SubCode,ULONG BufSize,PVOID Buffer)
{
}

//char buf[255]="Cooper has been shot, and lies nearly unconscious on the floor of his hotel room. Near death, Cooper is visited by an otherworldly Giant, who predicts three things: There is a smiling bag. The owls are not what they seem and without chemicals, he points.";
SYSTEMTIME stLocTime;


void menu()
{
	printf("-- version 1.0 (c)2008 (and.gus@hotmail.com)\n");
	printf("usage: mc-tos [-d destaddress] [-p port] [-t tos] [-l ttl]\n");
	printf("              [-w wait (ms)] [-s sender] [-r receiver] [-u update time]\n\n");
	printf("example sender: mc-tos -d 239.1.1.1 -p 9999 -t 46 -l 10 -w 2000 -s\n");
	printf("example receiver: mc-tos -d 239.1.1.1 -p 9999 -u -r\n");
	printf("default values: mc-tos -s (sender) or mc-tos -r (receiver)\n");
	exit (-1);
}

int main(int argc, char *argv[])
{
	// Default values
	int ttlValue=1;					// TTL
	int dscpValue=32;				// DiffServ
	int destPort=1234;				// UDP Port
	int	waitValue=500;				// Pause
	int updateTime=0;				// Update time on receiver on/off
	PCHAR destAddress="239.1.1.1";	// Multicast Address
	char val;						// Send or Receive
	int onscreen=1;					// On screen ?
	int counter=0;					// Packet count

	if(argc == 1)
	{
		menu();
	}

	for (int i=1; i<argc; i++)
	{
		if ((argv[i][0] == '-') || (argv[i][0] == '/') ) 
		{
			switch(tolower(argv[i][1]))
			{
				case 'd' :	// DestAddress
					destAddress = &argv[i][3];
				break;
				case 'l' :	// TTL
					ttlValue = atoi(&argv[i][3]);
				break;
				case 'p' :	// Port
					destPort = atoi(&argv[i][3]);
				break;
				case 't' :	// TOS
					dscpValue = atoi(&argv[i][3]);
				break;
				case 'w' :	// Pause between sending
					waitValue = atoi(&argv[i][3]);
				break;
				case 's' :	// Sender
					val='s';
				break;
				case 'r' :	// Receiver
					val='r';
				break;
				case 'q' :	// on screen on/off
					onscreen=0;
				break;
				case 'u' :	// Update time on Receiver
					updateTime=1;
				break;
				case 'h' :	// Help
					menu();
				break;
				case 'v' :
					printf("mc-tos version: 1.0\n");
					exit (-1);
				break;
				default:
					printf("Unknown options flag %s\n", argv[i]);
					exit (-1);
				break;
			}
		}
		//else
		//{
		//	//printf("  unknown option %s\n", argv[i]);
		//}
	}
	if (val == 'r')
		goto x;

	///////////////////////////////////////////////////////////////////////////////
	/// QOS STUFF
	///////////////////////////////////////////////////////////////////////////////
	HANDLE ClientHandle;
	HANDLE IfcHandle;
	HANDLE flowHandle;
	HANDLE FilterHandle;

	// Initialize call back functions. We dont' need these, but is required for TcRegisterClient API
	TCI_CLIENT_FUNC_LIST QoSFunctions;
	QoSFunctions.ClAddFlowCompleteHandler = NULL;
	QoSFunctions.ClDeleteFlowCompleteHandler = NULL;
	QoSFunctions.ClModifyFlowCompleteHandler = NULL;
	QoSFunctions.ClNotifyHandler = (TCI_NOTIFY_HANDLER)MyClNotifyHandler;

	// Register the client with Traffic control interface.
	long result = TcRegisterClient(CURRENT_TCI_VERSION, NULL,&QoSFunctions,&ClientHandle);

	//printf("TcRegisterClient returned %d 0x%x\n", result, result);

	if (result == NO_ERROR)
	{
		TC_IFC_DESCRIPTOR InterfaceBuffer[12];
		PTC_IFC_DESCRIPTOR pInterfaceBuffer = &InterfaceBuffer[0];
		ULONG BufferSize = 12 * sizeof(TC_IFC_DESCRIPTOR);

		// Find traffic control enabled interfaces on the cmachine
		result = TcEnumerateInterfaces(ClientHandle, &BufferSize, pInterfaceBuffer);
		if (result == NO_ERROR)
		{
			if (BufferSize == 0)
			{
				printf("no traffic control interfaces are available\n");
			}

			// The TcOpenInterface function identifies and opens an interface based on its text string, which is available from a call to TcEnumerateInterfaces
			result = TcOpenInterface(InterfaceBuffer[0].pInterfaceName, ClientHandle, NULL, &IfcHandle );
			if (result == NO_ERROR)
			{
				// Creating of Traffic flow headers
				int curSize = sizeof(TC_GEN_FLOW )+ sizeof(QOS_DS_CLASS) + sizeof(QOS_OBJECT_HDR);
				char *bufFlow = new char[curSize];
				PTC_GEN_FLOW newFlow = (PTC_GEN_FLOW) bufFlow;

				// Create the new temp flow.
				LPQOS_OBJECT_HDR objHdr = NULL;

				//Set the Flow Spec
				newFlow->ReceivingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.Latency = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.ServiceType = SERVICETYPE_GUARANTEED;	//SERVICETYPE_CONTROLLEDLOAD;
				newFlow->ReceivingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;
				newFlow->ReceivingFlowspec.TokenRate = QOS_NOT_SPECIFIED;			//10000;

				newFlow->SendingFlowspec.DelayVariation = QOS_NOT_SPECIFIED;
				newFlow->SendingFlowspec.Latency = QOS_NOT_SPECIFIED;
				newFlow->SendingFlowspec.MaxSduSize = QOS_NOT_SPECIFIED;			//344;
				newFlow->SendingFlowspec.MinimumPolicedSize = QOS_NOT_SPECIFIED;	//12;
				newFlow->SendingFlowspec.PeakBandwidth = QOS_NOT_SPECIFIED;			//32000;
				newFlow->SendingFlowspec.ServiceType = SERVICETYPE_GUARANTEED;		//SERVICETYPE_CONTROLLEDLOAD;
				newFlow->SendingFlowspec.TokenBucketSize = QOS_NOT_SPECIFIED;		//680;
				newFlow->SendingFlowspec.TokenRate = 10000;							//a guess - set to MTU
				newFlow->TcObjectsLength = sizeof QOS_DS_CLASS + sizeof QOS_OBJECT_HDR;

				// Set DSCP.
				LPQOS_DS_CLASS pQOSClass = (LPQOS_DS_CLASS)(&(newFlow->TcObjects[0]));
				pQOSClass->ObjectHdr.ObjectType = QOS_OBJECT_DS_CLASS;
				pQOSClass->ObjectHdr.ObjectLength = sizeof QOS_DS_CLASS;
				pQOSClass->DSField = dscpValue;
				//printf("%d",pQOSClass->DSField);
				//printf("\nSuccess: Setting TOS value - %d", pQOSClass->DSField);
				
				objHdr = (LPQOS_OBJECT_HDR)((char *)&(newFlow->TcObjects[0]) + sizeof(QOS_DS_CLASS));

				// Set the end of the list
				objHdr->ObjectType = QOS_OBJECT_END_OF_LIST;
				objHdr->ObjectLength = sizeof QOS_OBJECT_HDR;

				// adds a new flow on the specified interface. The flow is like a channel, that determines how to shape the traffic if it tunnelled through it.
				int retCode = TcAddFlow(IfcHandle, NULL, 0, newFlow, &flowHandle);

				// Create a filter. All traffic matching the filter will pass through the flow
				TC_GEN_FILTER GenericFilter;
				IP_PATTERN Pattern, Mask;
				memset(&Pattern,0,sizeof(IP_PATTERN));
				memset(&Mask,0,sizeof(IP_PATTERN));

				GenericFilter.AddressType = NDIS_PROTOCOL_ID_TCP_IP;	// Protocol stack is TCP/IP
				GenericFilter.PatternSize = sizeof(IP_PATTERN);
				GenericFilter.Pattern = &Pattern;						// pattern to match, defined below
				GenericFilter.Mask = &Mask;

				// Filter pattern. This is where you can tweak the flow filter parameters configured to match on all outgoing UDP packets.
				Pattern.Reserved1 = 0;
				Pattern.Reserved2 = 0;
				Pattern.SrcAddr = 0;									//inet_addr("10.213.24.82");
				Pattern.DstAddr = inet_addr(destAddress);
				Pattern.tcSrcPort = 0;
				Pattern.tcDstPort = 0;									//5000;
				Pattern.ProtocolId = IPPROTO_UDP;						//UDP
				Pattern.Reserved3[0] = 0;
				Pattern.Reserved3[1] = 0;
				Pattern.Reserved3[2] = 0;

				// Patterns mask
				Mask.Reserved1 = 0;
				Mask.Reserved2 = 0;
				Mask.SrcAddr = 0;										//0xFFFFFFFF;
				Mask.DstAddr = 0xFFFFFFFF;
				Mask.tcSrcPort = 0;										//0xFFFF;
				Mask.tcDstPort = 0;										//0xFFFF;
				Mask.ProtocolId = 0;									//0xFF;
				Mask.Reserved3[0] = 0;
				Mask.Reserved3[1] = 0;
				Mask.Reserved3[2] = 0;

				// Add filter to the flow. This point onwards all traffic matching the filter, will pass through the flow.
				retCode = TcAddFilter(flowHandle, &GenericFilter, &FilterHandle);

				if(retCode != NO_ERROR)
				{
					switch(retCode)
					{
						case ERROR_SIGNAL_PENDING:
							printf("The function is being executed asynchronously; the client will be called back through the client-exposed ClAddFlowComplete function when the flow has been added or when the process has been completed.\n");
						break;
						case ERROR_INVALID_HANDLE:
							printf("The interface handle is invalid.\n");
						break;
						case ERROR_NOT_ENOUGH_MEMORY:
							printf("The system is out of memory.\n");
						break;
						case ERROR_INVALID_PARAMETER:
							printf("A parameter is invalid.\n");
						break;
						case ERROR_INVALID_SERVICE_TYPE:
							printf("An unspecified or bad INTSERV service type has been provided.\n");
						break;
						case ERROR_INVALID_TOKEN_RATE:
							printf("An unspecified or bad TOKENRATE value has been provided.\n");
						break;
						case ERROR_INVALID_PEAK_RATE:
							printf("The PEAKBANDWIDTH value is invalid.\n");
						break;
						case ERROR_INVALID_SD_MODE:
							printf("The SHAPEDISCARDMODE is invalid.\n");
						break;
						case ERROR_INVALID_QOS_PRIORITY:
							printf("The priority value is invalid.\n");
						break;
						case ERROR_INVALID_TRAFFIC_CLASS:
							printf("The traffic class value is invalid.\n");
						break;
						case ERROR_NO_SYSTEM_RESOURCES:
							printf("There are not enough resources to accommodate the requested flow.\n");
						break;
						case ERROR_TC_OBJECT_LENGTH_INVALID:
							printf("Bad length specified for the TC objects.\n");
						break;
						case ERROR_INVALID_DIFFSERV_FLOW:
							printf("Applies to Diffserv flows. Indicates that the QOS_DIFFSERV object was passed with an invalid parameter.\n");
						break;
						case ERROR_DS_MAPPING_EXISTS:
							printf("Applies to Diffserv flows. Indicates that the QOS_DIFFSERV_RULE specified in TC_GEN_FLOW already applies to an existing flow on the interface.\n");
						break;
						case ERROR_INVALID_SHAPE_RATE:
							printf("The QOS_SHAPING_RATE object was passed with an invalid ShapeRate.\n");
						break;
						case ERROR_INVALID_DS_CLASS:
							printf("The QOS_DS_CLASS is invalid.\n");
						break;
						case ERROR_NETWORK_UNREACHABLE:
							printf("The network cable is not plugged into the adapter.\n");
						break;
					}
					printf("TcAddFlow: ERROR: %lu\n",retCode);
					printf("Handle was %d\n", (int)IfcHandle);
					exit (-1);
				}
			}
			else
			{
				printf("TcOpenInterface returned 0x%x\n", result);
				exit (-1);
			}
		}
		else
		{
			switch(result)
			{
				case NO_ERROR:
					printf("The function executed without errors.\n");
				break;
				case ERROR_INVALID_HANDLE:
					printf("The client handle is invalid.\n");
				break;
				case ERROR_INVALID_PARAMETER:
					printf("One of the parameters is NULL.\n");
				break;
				case ERROR_INSUFFICIENT_BUFFER:
					printf("The buffer is too small to enumerate all interfaces. If this error is returned, the correct (required) size of the buffer is passed back in pBufferSize.\n");
					printf("Size required: %d\n", BufferSize);
				break;
				case ERROR_NOT_ENOUGH_MEMORY:
					printf("The system is out of memory.\n");
				break;
			}
			printf("TcEnumerateInterfaces returned 0x%x\n", result);
			exit (-1);
		}
	}
	else
	{
		printf("TcRegisterClient returned 0x%x\n", result);
		exit (-1);
	}

	///////////////////////////////////////////////////////////////////////////////
	/// INIT WINSOCK & SEND/RECEIVE TRAFFIC
	///////////////////////////////////////////////////////////////////////////////
	x:
	int err;
	WSADATA WSAData;

	err = WSAStartup(MAKEWORD(2,2), &WSAData);
	if (err)
	{
		printf("WSAStartup Failed (%d) Exiting\n", err);
		exit(err);
	}

	// Create a normal UDP socket for sending data
	SOCKET sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sd == INVALID_SOCKET)
	{
		err = GetLastError();
		printf("\"Socket\" failed with error %d\n", err);
		exit (-1);
	}

	///////////////////////////////////////////////////////////////////////////////
	/// My stuff...
	///////////////////////////////////////////////////////////////////////////////
	switch(val)
	{
		case('s'):

		sockaddr_in send_addr;
		memset (&send_addr, 0, sizeof(SOCKADDR_IN));
		send_addr.sin_family = AF_INET;
		send_addr.sin_addr.s_addr = inet_addr (destAddress);					// this is our test destination IP
		send_addr.sin_port = htons(destPort);

		// Set the Time-to-Live of the multicast
		if(setsockopt(sd,IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttlValue,sizeof (int)) == SOCKET_ERROR)
		{
			printf("Error: Setting TTL value: error - %d\n",WSAGetLastError());
			closesocket (sd);
			exit (-1);
		}
		//printf("\nSuccess: Setting TTL value - %d\n", ttlValue);

		printf("\ndestAddress = %s\n", destAddress);
		printf("destPort    = %d\n", destPort);
		printf("dscpValue   = %d\n", dscpValue);
		printf("ttlValue    = %d\n", ttlValue);
		printf("waitValue   = %d ms\n", waitValue);

		// Get System Local Time and send a Multicast packet
		printf("\nSending...\n");

		for(;;)
		{
			GetLocalTime (&stLocTime);

			if(sendto(sd,(char *)&stLocTime,sizeof(stLocTime), MSG_DONTROUTE, (sockaddr*)&send_addr, sizeof(send_addr))==INVALID_SOCKET)
			{
				printf("\nError in Sending data on the socket - %d", WSAGetLastError());
				WSACleanup();
				exit (-1);
			}
			if(onscreen)
			{
				printf("Sent Time %02d:%02d:%02d:%03d on %02d-%02d-%02d to: %s:%d tos:%d ttl:%d seq:%d\n",
				stLocTime.wHour,
				stLocTime.wMinute,
				stLocTime.wSecond,
				stLocTime.wMilliseconds,
				stLocTime.wMonth,
				stLocTime.wDay,
				stLocTime.wYear,
				destAddress,
				destPort,
				dscpValue,
				ttlValue,
				counter++);
				Sleep(waitValue);
			}
		}
		closesocket(sd);
		WSACleanup();
		exit (-1);
		break;
		///////////////////////////////////////////////////////////////////////////////
		case('r'):

		char time_message[100];					// ASCII string 
		int iRecvLen;							// Length of receive_addr
		SYSTEMTIME *lpstLocTime;
		lpstLocTime = (SYSTEMTIME *)(time_message);

		sockaddr_in receive_addr;		// Local socket's address. Holds the source address upon recvfrom function returns
		receive_addr.sin_family = AF_INET;
		receive_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		receive_addr.sin_port = htons(destPort);
		iRecvLen = sizeof (receive_addr);

		// Binding the socket to the address. Only necessary on receiver
		if(bind(sd,(sockaddr*)&receive_addr, sizeof(receive_addr))==SOCKET_ERROR)
		{
			printf("\nError: Binding the socket - %d", WSAGetLastError());
			WSACleanup();
			exit (-1);
		}
		printf("\nSuccess: Binding the socket - %d", WSAGetLastError());

		ip_mreq req;
		req.imr_multiaddr.s_addr=inet_addr (destAddress);
		req.imr_interface.s_addr =INADDR_ANY;
		// Joining a multicast group
		if(setsockopt(sd,IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&req, sizeof(req)) == -1)
		{
			printf("\nError: Joining Multicast Group - %d", WSAGetLastError());
			WSACleanup();
			exit (-1);
		}
		printf("\nSuccess: Joining Multicast Group - %d\n", WSAGetLastError());

		printf("\ndestAddress = %s\n", destAddress);
		printf("destPort    = %d\n", destPort);
		if(updateTime)
			printf("Update      = TRUE\n");
		else
			printf("Update      = FALSE\n");

		// Receive data from the multicasting group server.
		printf("\nReceiving...\n");

		for(;;)
		{
			if (recvfrom (sd,time_message, 100, 0, (struct sockaddr FAR *)&receive_addr, &iRecvLen) == SOCKET_ERROR)
			{
				printf ("\nrecvfrom failed: %d",WSAGetLastError ());
				WSACleanup();
				exit (-1);
			}
			else
			{
				printf("Received Time %02d:%02d:%02d:%03d on %02d-%02d-%d seq:%d\n",
					lpstLocTime->wHour,
					lpstLocTime->wMinute,
					lpstLocTime->wSecond,
					lpstLocTime->wMilliseconds,
					lpstLocTime->wMonth,
					lpstLocTime->wDay,
					lpstLocTime->wYear,
					counter++);
				if(updateTime)	// Update local time
				{
					SetLocalTime(lpstLocTime);
					printf("Time updated..\n");
				}
			}
		}

		// Dropping membership from the Mcast group
		if(setsockopt(sd, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char*)&req, sizeof(req)) == -1)
		{
			printf("\nError: Leaving Multicast Group - %d", WSAGetLastError());
			WSACleanup();
			exit (-1);
		}
		printf("\nSuccess: Leaving Multicast Group - %d", WSAGetLastError());

		closesocket(sd);
		WSACleanup();
		exit (-1);
		break;
	}
	// End My stuff

	///////////////////////////////////////////////////////////////////////////////
	/// CLEANUP
	///////////////////////////////////////////////////////////////////////////////
	result = TcDeleteFilter(FilterHandle);
	if (result != NO_ERROR)
	{
		printf("Error deleting filter\n");
	}
	// Delete flow associated with the network interface
	result = TcDeleteFlow(flowHandle);
	if (result != NO_ERROR)
	{
		printf("Error deleting flow\n");
	}

	// Close the Traffic control interface opened earlier.
	result = TcCloseInterface(IfcHandle);
	if (result != NO_ERROR)
	{
		switch(result)
		{
		case ERROR_INVALID_HANDLE:
			printf("The interface handle is invalid.\n");
		break;
		case ERROR_TC_SUPPORTED_OBJECTS_EXIST:
			printf("Not all flows have been deleted for this interface.");
		}
		printf("Error closing interface\n");
	}

	// Unregister the client.
	result = TcDeregisterClient(ClientHandle);
	if (result != NO_ERROR)
	{
		switch(result)
		{
		case ERROR_INVALID_HANDLE:
			printf("Invalid interface handle, or the handle was set to NULL.\n");
		break;
		case ERROR_TC_SUPPORTED_OBJECTS_EXIST:
			printf("Interfaces are still open for this client. all interfaces must be closed to deregister a client.\n");
		}
		printf("Error degresitering client\n");
	}
	// shutdown the socket
	shutdown(sd, SD_BOTH);
	closesocket(sd);
	WSACleanup();
	return 0;
}
